/* Copyright (C) 2002 Ricoh Co.,Ltd. All Rights Reserved.

   This file is part of Ghostscript.

   GNU Ghostscript is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
   to anyone for the consequences of using it or for whether it serves any
   particular purpose or works at all, unless he says so in writing.  Refer
   to the GNU General Public License for full details.

   Everyone is granted permission to copy, modify and redistribute
   this software, but only under the conditions described in the GNU
   General Public License.  A copy of this license is supposed to have been
   given to you along with this software so you can know your rights and
   responsibilities.  It should be in a file named COPYING.  Among other
   things, the copyright notice and this notice must be preserved on all
   copies.
 */

/* gdevrpdl4.c */
/* Ricoh RPDL IV driver */

#include "gdevprn.h"

#define DPI 400	 

/*
 * ghostscriptǻѲǽʥץ
 */
static struct _rpdl_cont{
	int BW;		/* block Width  (byte:8dot)	*/
	int BH;		/* block height (line)		*/
	int Tray;	/* select tray no. */
	bool manFeed;	/* select manualfeed */
	int FrmOv;
	int PRange;
	int LB;		/* ; */
	int UB;		/* ; */
} rpdl_cont =
{			/* default params	*/
	1,		/* block width  (byte)	*/
	24,		/* block height (line)	*/
	0,		/* tray no. */
	0,		/* manualfeed */
	0,
	0,
	0,
	0,
};

/*
 * ݻ빽¤
 */
typedef struct {
	gx_device_printer *pdev;
	FILE	*pstream;

	int blw;
	int blh;

	int x_dpi;
	int LINE_SIZE;
	int h;
	int w;
} RpdlPageCont;

/* The device descriptors */
private dev_proc_print_page(paper_size_set);
private dev_proc_get_params(rpdl_get_params);
private dev_proc_put_params(rpdl_put_params);
private void rpdl_page_cont_init(gx_device_printer *pdev, FILE *pstream, RpdlPageCont *cont);

/*
 * ǥХ
 */
private gx_device_procs prn_rpdl_procs =
prn_params_procs(gdev_prn_open, gdev_prn_output_page, gdev_prn_close,
		 rpdl_get_params, rpdl_put_params);
gx_device_printer far_data gs_rpdl4_device =
prn_device(prn_rpdl_procs, "rpdl4",
           DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS,
           DPI, DPI,		/* x_dpi, y_dpi */
           0.20, 0.20, 0.20, 0.20,	/* margins */
           1, paper_size_set);

#define ppdev ((gx_device_printer *)pdev)

private int
rpdl_get_params(gx_device *pdev, gs_param_list *plist)
{
	int code;

	if(ppdev->Duplex_set < 0)
		ppdev->Duplex_set = 0;

	if((code = gdev_prn_get_params(pdev, plist)) < 0 ||
	   (code = param_write_int(plist, "BW", &rpdl_cont.BW)) < 0 ||
	   (code = param_write_int(plist, "BH", &rpdl_cont.BH)) < 0 ||
	   (code = param_write_bool(plist, "ManFeed", &rpdl_cont.manFeed)) < 0 ||
	   (code = param_write_int(plist, "Tray", &rpdl_cont.Tray)) < 0 ||
	   (code = param_write_int(plist, "FrmOv", &rpdl_cont.FrmOv)) < 0 ||
	   (code = param_write_int(plist, "PRange", &rpdl_cont.PRange)) < 0 ||
	   (code = param_write_int(plist, "LB", &rpdl_cont.LB)) < 0 ||
	   (code = param_write_int(plist, "UB", &rpdl_cont.UB)) < 0
	   )
	   return code;
	 return code;
}

/* put properties. */
private int
rpdl_put_params(gx_device *pdev, gs_param_list *plist)
{
	param_read_int(plist, "BW", &rpdl_cont.BW);
	param_read_int(plist, "BH", &rpdl_cont.BH);
	param_read_bool(plist, "ManFeed", &rpdl_cont.manFeed);
	param_read_int(plist, "Tray", &rpdl_cont.Tray);
	param_read_int(plist, "FrmOv", &rpdl_cont.FrmOv);
	param_read_int(plist, "PRange", &rpdl_cont.PRange);
	param_read_int(plist, "LB", &rpdl_cont.LB);
	param_read_int(plist, "UB", &rpdl_cont.UB);
	gdev_prn_put_params(pdev, plist);
	return 0;
}

/*
 * 
 */
private void rpdl_page_cont_init(gx_device_printer *pdev, FILE *pstream,
				 RpdlPageCont *cont)
{
	cont->pdev = pdev;
	cont->pstream = pstream;

	cont->blw = rpdl_cont.BW;
	if(cont->blw < 1) cont->blw = 1;
	cont->blh = rpdl_cont.BH;
	if(cont->blh < 1) cont->blh = 24;

	cont->x_dpi = pdev->x_pixels_per_inch;
	cont->LINE_SIZE = 
		(pdev->x_pixels_per_inch * 2970 / 254 / 8 + 1);
	cont->h = pdev->height;
	cont->w = gdev_prn_raster(pdev);

}

#define	SUB 0x1a
#define ESC 0x1b
#define	DC2 0x12
#define	EM  0x19
#define SO  0x0e

int nPrintCount;
int nPrintBlock;
int nTotalBlock;


/*
  Send the block to the printer. 
    pstream: 	output file 
    bw: 	width in byte
    bh: 	height in dot
    x : 	start x position in byte
    y : 	start y position in dot
    row_buf: 	data buffer
    w:		line length in byte
*/
private int
rpdl_print_block(RpdlPageCont *cont, int bw, int bh, int x, int y, char* row_buf)
{
	char block_data[cont->blh * cont->LINE_SIZE];
	char comp_data[cont->blh * cont->LINE_SIZE * 2];
	int total_size = bw * bh;
	int compress = 0;
	char *bgn_data;
	int c, count, i, j;
	char *rp, *cp, *sp;

	/* compress */
	sp = block_data;
	cp = comp_data;
	c = *(row_buf + x) + 1;
	count = 1;
	for(j = 0; j < bh; j++, sp += bw) {
		rp = row_buf + cont->w * j + x;
		memcpy(sp, rp, bw);
		for(i = 0; i < bw; i++, rp++) {
			if(*rp == c) {
				count++;
				if(count == 258) {
					*cp++ = c;
					*cp++ = (char)255;
					*cp++ = c;
					count = 1;
				}
			} else {
				if(count > 1) {
					*cp++ = c;
					*cp++ = count - 2;
					count = 1;
				}
				*cp++ = c = *rp;
			}
        	}
	}
	if(count > 1) {
		*cp++ = c;
		*cp++ = count - 2;
	}

	/* check whether use compress data or not */
	if(total_size > cp - comp_data) {
		bgn_data = comp_data;
		total_size = cp - comp_data;
		compress = 4;
	} else {
		bgn_data = block_data;
		total_size = bw * bh;
		compress = 0;
	}

	/* move */
	fprintf(cont->pstream, "%c%cG3,%d,%d,1,%d,%d,%d,%d@", ESC, DC2, 
		bw*8,		/* dots per line */
		bh,		/* lines */
		compress,	/* compress type */
		x*8,		/* start position */
		y,		/* start line position */
		total_size	/* data size */
	);

	/* send the block */
	fwrite(bgn_data, sizeof(char), total_size, cont->pstream);

	return 0;
}

/* Send the page to the printer.  */
private int
rpdl_print_page(RpdlPageCont *cont)
{

	/* devide all lines in several blocks */
	int bh, bw, x, y, first_x, block_width;

	char row_buf[cont->blh * cont->LINE_SIZE];

	nPrintCount = nPrintBlock = nTotalBlock = 0;

	/* print blocks */
	for(y = 0; y < cont->h; y += cont->blh) {
		bh = cont->h - y;
		if(bh > cont->blh) 
			bh = cont->blh;

		gdev_prn_copy_scan_lines(cont->pdev, y, row_buf, bh * cont->w);

		first_x = -1;
		block_width = 0;
		for(x = 0; x < cont->w; x += cont->blw) {
			bw = cont->w - x;
			if(bw > cont->blw)
				bw = cont->blw;

			nTotalBlock++;
			/* skip white block */
			{
				int i, j;
				for(j = 0; j < bh; j++) {
					char* rp = row_buf + j * cont->w + x;
					for(i = 0; i < bw; i++, rp++) {
						if(*rp) {
							goto has_data;
						}
					}
				}
				if(block_width) {
					rpdl_print_block(cont,
						block_width, bh,
						first_x, y,
						row_buf);
					nPrintCount++;
					block_width = 0;
					first_x = -1;
				} 
				continue; 
			}
has_data:
			if(first_x == -1) {
				first_x = x;
			}
			block_width += bw;
			nPrintBlock++;

		}
		if(block_width) {
			rpdl_print_block(cont,
				block_width, bh,
				first_x, y,
				row_buf);
			nPrintCount++;
		}
	}

	/* print page */
	fputc(0x0C, cont->pstream);

	return 0;
}

private int
rpdl_print_init(gx_device_printer *pdev, FILE *pstream, const char* paper_size)
{
	RpdlPageCont cont;
	int dpi;

	dpi = pdev->x_pixels_per_inch;

	if(pdev->PageCount == 0){
	  /* initialize */
	  fprintf(pstream, "%c%c!1@R00%c ", ESC, DC2, ESC);
	  fprintf(pstream, "%c4", ESC);
	  fprintf(pstream, "%c%cI", ESC, SUB);
	  fprintf(pstream, "%c%cYP,1 ", ESC, DC2);

	  switch(dpi) {
	  case 400:		
	    fprintf(pstream, "%c%cYW,1 ", ESC, DC2);
	    fprintf(pstream, "%c%cYI,7 ", ESC, DC2);
	    fprintf(pstream, "%c%cYA04,1 ", ESC, DC2);
	    fprintf(pstream, "%c%c#2 ", ESC, DC2);
	    fprintf(pstream, "%c%cQ4 ", ESC, DC2);
	    break;
	  
	  case 600:
	    fprintf(pstream, "%c%cYW,3 ", ESC, DC2);
	    fprintf(pstream, "%c%cYI,8 ", ESC, DC2);
	    fprintf(pstream, "%c%cYA04,3 ", ESC, DC2);
	    fprintf(pstream, "%c%c#4 ", ESC, DC2);
	    fprintf(pstream, "%c%cQ5 ", ESC, DC2);
	    break;

	  case 1200:
	    fprintf(pstream, "%c%cYW,4 ", ESC, DC2);
	    fprintf(pstream, "%c%cYI,9 ", ESC, DC2);
	    fprintf(pstream, "%c%cYA04,4 ", ESC, DC2);
	    fprintf(pstream, "%c%c#5 ", ESC, DC2);
	    fprintf(pstream, "%c%cQ6 ", ESC, DC2);
	    break;

	  default: /* same as 400dpi */
	    fprintf(pstream, "%c%cYW,1 ", ESC, DC2);
	    fprintf(pstream, "%c%cYI,7 ", ESC, DC2);
	    fprintf(pstream, "%c%cYA04,1 ", ESC, DC2);
	    fprintf(pstream, "%c%c#,2 ", ESC, DC2);
	    fprintf(pstream, "%c%cQ4 ", ESC, DC2);
	    break;
	  }

	  fprintf(pstream, "%c%cYJ,1 ", ESC, DC2);
	  fprintf(pstream, "%c%cYQ,2 ", ESC, DC2);

	  if(rpdl_cont.PRange){
	    switch(rpdl_cont.PRange){
	    case 1:
	      fprintf(pstream, "%c%cYB,1 ", ESC, DC2);
	      break;
	    case 2:
	      fprintf(pstream, "%c%cYB,2 ", ESC, DC2);
	      break;
	    case 3:
	      fprintf(pstream, "%c%cYB,2 ", ESC, DC2);
	      break;
	    default:
	      fprintf(pstream, "%c%cYB,1 ", ESC, DC2);
	      break;
	    }
	  } else {
	    fprintf(pstream, "%c%cYB,1 ", ESC, DC2);
	  }

	  fprintf(pstream, "%c%cYA02,2 ", ESC, DC2);

	  if(pdev->Duplex_set > 0){
	    if(pdev->Duplex){
	      fprintf(pstream, "%c%cYA06,2 ", ESC, DC2);
	      fprintf(pstream, "%c%cYA03,2 ", ESC, DC2);
	    }else{
	      fprintf(pstream, "%c%cYA06,1 ", ESC, DC2);
	    }
	  }
	
	  fprintf(pstream, "%c%cYM,1 ", ESC, DC2);

	  fprintf(pstream, "%c%c?M100,0%c ", ESC, DC2, ESC);

	  if(rpdl_cont.manFeed > 0){
	    fprintf(pstream, "%c%cZ1 ", ESC, DC2);
	    fprintf(pstream, "%c%cM", ESC, EM);
	    fprintf(pstream, "%c%c52@%.3s.Plain%c ", ESC, DC2, paper_size, ESC);
	    fprintf(pstream, "%c%cM", ESC, EM);
	  } else {
	    switch(rpdl_cont.Tray){
	    case 0:
	      fprintf(pstream, "%c%cZ2 ", ESC, DC2);
	      fprintf(pstream, "%c%c53@%.2sX.Plain%c ", ESC, DC2, paper_size, ESC);
	      break;
	    case 1:
	      fprintf(pstream, "%c%cZ1 ", ESC, DC2);
	      fprintf(pstream, "%c%c2", ESC, EM);
	      break;
	    case 2:
	      fprintf(pstream, "%c%cZ1 ", ESC, DC2);
	      fprintf(pstream, "%c%c3", ESC, EM);
	      break;
	    case 3:
	      fprintf(pstream, "%c%cZ1 ", ESC, DC2);
	      fprintf(pstream, "%c%c4", ESC, EM);
	      break;
	    case 4:
	      fprintf(pstream, "%c%cZ1 ", ESC, DC2);
	      fprintf(pstream, "%c%c5", ESC, EM);
	      break;
	    case 5:
	      fprintf(pstream, "%c%cZ1 ", ESC, DC2);
	      fprintf(pstream, "%c%c6", ESC, EM);
	      break;
	    default:
	      fprintf(pstream, "%c%cZ2 ", ESC, DC2);
	      fprintf(pstream, "%c%c53@%.2sX.Plain%c ", ESC, DC2, paper_size, ESC);
	      break;
	    }
	  }

	  if(pdev->MediaSize[0] > pdev->MediaSize[1]){
	    fprintf(pstream, "%c%cD2 ", ESC, DC2);
	  }else{
	    fprintf(pstream, "%c%cD1 ", ESC, DC2);
	  }

	  fprintf(pstream, "%c%c", ESC, SO);
	  fprintf(pstream, "%c%cN1 ", ESC, DC2);

	  if(rpdl_cont.LB > 0 || rpdl_cont.UB > 0){
	    fprintf(pstream, "%c%c?Y0:0,L:%d,K:%d%c ", ESC, DC2, rpdl_cont.UB, rpdl_cont.LB, ESC);
	  }
	}

	if(rpdl_cont.FrmOv){
	  fprintf(pstream, "%c%cU%02d ", ESC, DC2, rpdl_cont.FrmOv);
	}

	rpdl_page_cont_init(pdev, pstream, &cont);
	return rpdl_print_page(&cont);
}

/*
 * ѻ極ݻ빽¤
 */
typedef struct {
	int width;
	int height;
	int count;
	char name[3];
} RpdlPaperTable;

static RpdlPaperTable rpdlPaperTable[] =
{
  { 842, 1190,  1, "A3R"}, /* a3 */
  { 595,  842,  2, "A4R"}, /* a4 */
  { 842,  595,  3, "A4 "}, /* a4 */
  { 420,  595,  4, "A5R"}, /* a5 */
  { 595,  420,  5, "A5 "}, /* a5 */
  { 297,  420,  6, "A6R"}, /* a6 */
  { 283,  419,  7, "PCR"}, /* hagaki */
  { 728, 1032,  7, "B4R"}, /* b4 */
  { 515,  728,  8, "B5R"}, /* b5 */
  { 729,  516,  9, "B5 "}, /* b5 */
  { 364,  516, 10, "B6R"}, /* b6 */
  {   0,    0, -1, "A4 "}  /* error */
};

private int
paper_size_set(gx_device_printer *pdev, FILE *pstream)
{
	int width, height, t_width;
	char paper_size[4];
	RpdlPaperTable *pt;

	width  = pdev->MediaSize[0];
	height = pdev->MediaSize[1];

	if(width > height){
		t_width = width;
		width = height;
		height = t_width;
	}

	for(pt = rpdlPaperTable; pt->count > 0; pt++)
		if(pt->width == width && pt->height == height)
			break;

	memcpy(paper_size, pt->name, 3);

	return rpdl_print_init(pdev, pstream, paper_size);
}
